from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from math import log
from typing import Dict, Generic, List, Optional, Tuple, Type

import imm
from imm import (
    DP,
    HMM,
    CData,
    MuteState,
    Path,
    Sequence,
    State,
    Step,
    lprob_add,
    lprob_is_zero,
    lprob_zero,
)

from .typing import MutableResults, MutableState, TState

__all__ = [
    "AltModel",
    "EntryDistr",
    "Node",
    "NullModel",
    "SpecialNode",
    "SpecialTransitions",
    "Transitions",
]


class EntryDistr(Enum):
    UNIFORM = 1
    OCCUPANCY = 2


@dataclass
class Transitions:
    MM: float = lprob_zero()
    MI: float = lprob_zero()
    MD: float = lprob_zero()
    IM: float = lprob_zero()
    II: float = lprob_zero()
    DM: float = lprob_zero()
    DD: float = lprob_zero()

    def normalize(self):
        m_norm: float = lprob_add(lprob_add(self.MM, self.MI), self.MD)
        self.MM -= m_norm
        self.MI -= m_norm
        self.MD -= m_norm

        i_norm: float = lprob_add(self.IM, self.II)
        self.IM -= i_norm
        self.II -= i_norm

        d_norm: float = lprob_add(self.DM, self.DD)
        self.DM -= d_norm
        self.DD -= d_norm


@dataclass
class SpecialTransitions:
    NN: float = 0.0
    NB: float = 0.0
    EC: float = 0.0
    CC: float = 0.0
    CT: float = 0.0
    EJ: float = 0.0
    JJ: float = 0.0
    JB: float = 0.0
    RR: float = 0.0
    BM: float = 0.0
    ME: float = 0.0


class Node(Generic[TState]):
    def __init__(self, M: TState, I: TState, D: MuteState):
        self._M = M
        self._I = I
        self._D = D

    @property
    def M(self) -> TState:
        return self._M

    @property
    def I(self) -> TState:
        return self._I

    @property
    def D(self) -> MuteState:
        return self._D

    def states(self) -> List[MutableState[TState]]:
        return [self._M, self._I, self._D]


class SpecialNode(Generic[TState]):
    def __init__(
        self,
        S: MuteState,
        N: TState,
        B: MuteState,
        E: MuteState,
        J: TState,
        C: TState,
        T: MuteState,
    ):
        self._S = S
        self._N = N
        self._B = B
        self._E = E
        self._J = J
        self._C = C
        self._T = T

    @property
    def S(self) -> MuteState:
        return self._S

    @property
    def N(self) -> TState:
        return self._N

    @property
    def B(self) -> MuteState:
        return self._B

    @property
    def E(self) -> MuteState:
        return self._E

    @property
    def J(self) -> TState:
        return self._J

    @property
    def C(self) -> TState:
        return self._C

    @property
    def T(self) -> MuteState:
        return self._T

    def states(self) -> List[MutableState[TState]]:
        return [self._S, self._N, self._B, self._E, self._J, self._C, self._T]


class NullModel(Generic[TState]):
    def __init__(self, hmm: HMM, state: TState):
        self._hmm = hmm
        self._state = state

    @classmethod
    def create(cls: Type[NullModel], state: TState) -> NullModel:
        hmm = HMM.create(state.alphabet)
        hmm.add_state(state, 0.0)
        return cls(hmm, state)

    @classmethod
    def create_from_hmm(cls: Type[NullModel], hmm: HMM) -> NullModel:
        states = hmm.states()
        if len(states) != 1:
            raise ValueError("Null HMM must have only one state.")
        return cls(hmm, list(states.values())[0])

    @property
    def hmm(self) -> HMM:
        return self._hmm

    @property
    def state(self) -> TState:
        return self._state

    def set_transition(self, lprob: float):
        self._hmm.set_transition(self.state, self.state, lprob)

    def likelihood(self, sequence: Sequence):
        steps = [Step.create(self.state, 1) for i in range(len(sequence))]
        path = Path.create(steps)
        return self._hmm.likelihood(sequence, path)

    def set_special_transitions(self, special_trans: SpecialTransitions):
        self.set_transition(special_trans.RR)

    def __str__(self):
        return f"{self._hmm}"


class AltModel(Generic[TState]):
    def __init__(
        self,
        special_node: SpecialNode,
        core_nodes: List[Node],
        states: Dict[CData, MutableState[TState]],
        hmm: HMM,
        dp: Optional[DP[TState]] = None,
    ):
        self._special_node = special_node
        self._core_nodes = core_nodes
        self._states: Dict[CData, MutableState[TState]] = states
        self._hmm = hmm
        self._dp: Optional[DP[TState]] = dp

    @classmethod
    def create(
        cls: Type[AltModel],
        special_node: SpecialNode,
        core_nodes: List[Node],
        core_trans: List[Transitions],
        entry_distr: EntryDistr,
    ) -> AltModel:

        states: Dict[CData, MutableState[TState]] = {}

        for node in core_nodes:
            for state in node.states():
                states[state.imm_state] = state

        for state in special_node.states():
            states[state.imm_state] = state

        hmm = HMM.create(special_node.S.alphabet)
        hmm.add_state(special_node.S, 0.0)
        hmm.add_state(special_node.N)
        hmm.add_state(special_node.B)
        hmm.add_state(special_node.E)
        hmm.add_state(special_node.J)
        hmm.add_state(special_node.C)
        hmm.add_state(special_node.T)

        for node in core_nodes:
            hmm.add_state(node.M)
            hmm.add_state(node.I)
            hmm.add_state(node.D)

        M1 = core_nodes[0].M
        hmm.set_transition(special_node.B, M1, core_trans[0].MM)

        for i, trans in enumerate(core_trans[1:-1]):
            prev = core_nodes[i]
            next = core_nodes[i + 1]
            hmm.set_transition(prev.M, prev.I, trans.MI)
            hmm.set_transition(prev.I, prev.I, trans.II)
            hmm.set_transition(prev.M, next.M, trans.MM)
            hmm.set_transition(prev.I, next.M, trans.IM)
            hmm.set_transition(prev.M, next.D, trans.MD)
            hmm.set_transition(prev.D, next.D, trans.DD)
            hmm.set_transition(prev.D, next.M, trans.DM)

        Mm = core_nodes[-1].M
        hmm.set_transition(Mm, special_node.E, core_trans[-1].MM)

        alt_model = cls(special_node, core_nodes, states, hmm)

        alt_model.set_entry_transitions(entry_distr, core_trans)
        alt_model.set_exit_transitions()

        return alt_model

    @classmethod
    def create_from_hmm(
        cls: Type[AltModel],
        special_node: SpecialNode,
        core_nodes: List[Node],
        hmm: HMM,
        dp: DP,
    ) -> AltModel:

        states: Dict[CData, MutableState[TState]] = {}

        for node in core_nodes:
            for state in node.states():
                states[state.imm_state] = state

        for state in special_node.states():
            states[state.imm_state] = state

        alt_model = cls(special_node, core_nodes, states, hmm, dp)

        # alt_model.set_entry_transitions(entry_distr, core_trans)
        # alt_model.set_exit_transitions()

        return alt_model

    @property
    def hmm(self) -> HMM:
        return self._hmm

    def view(self, core_model_only=False):
        from graphviz import Digraph

        dot = Digraph(comment="Profile")
        dot.attr(rankdir="LR")

        dot.node(self._special_node.B.name.decode())

        if not core_model_only:
            states = [self._special_node.S, self._special_node.N]
            for state in states:
                dot.node(state.name.decode())

        rank = 0
        for core in self._core_nodes:
            with dot.subgraph() as cluster:
                cluster.attr(rank=str(rank))
                cluster.node(core.M.name.decode(), shape="square")
                cluster.node(core.D.name.decode(), shape="circle")
                cluster.node(core.I.name.decode(), shape="diamond")
            rank += 1

        dot.node(self._special_node.E.name.decode())

        if not core_model_only:
            states = [self._special_node.C, self._special_node.T]
            for state in states:
                dot.node(state.name.decode())

            state = self._special_node.J
            dot.node(state.name.decode())

        for s0 in self._states.values():
            for s1 in self._states.values():
                t = self._hmm.transition(s0, s1)
                if lprob_is_zero(t):
                    continue
                label = f"{t:.4f}"
                special = self._is_special_state(s0) or self._is_special_state(s1)
                core_trans = self._is_core_state(s0) and self._is_core_state(s1)
                if core_model_only and not core_trans:
                    continue

                name0 = s0.name.decode()
                name1 = s1.name.decode()
                if special:
                    color = "lightgrey"
                else:
                    color = "black"
                dot.edge(name0, name1, label=label, color=color)

        dot.view()

    def _is_special_state(self, state: MutableState[TState]) -> bool:
        return state in self._special_node.states()

    def _is_core_state(self, state: MutableState[TState]) -> bool:
        B = self._special_node.B
        E = self._special_node.E
        return state not in self._special_node.states() or state in [B, E]

    def set_transition(self, a: State, b: State, lprob: float):
        # self._dp = None
        self._hmm.set_transition(a, b, lprob)

    def core_nodes(self) -> List[Node]:
        return self._core_nodes

    @property
    def special_node(self) -> SpecialNode:
        return self._special_node

    @property
    def core_length(self) -> int:
        return len(self._core_nodes)

    def viterbi(self, seq: Sequence, window_length: int = 0) -> MutableResults[TState]:
        if self._dp is None:
            self._dp = self._hmm.create_dp(self.special_node.T)

        return self._dp.viterbi(seq, window_length)

    def set_fragment_length(self, special_trans: SpecialTransitions):
        del special_trans
        raise RuntimeError("Fix this function to update dp.")
        # M = self.core_length
        # if M == 0:
        #     return

        # self._dp = None
        # B = self.special_node.B
        # E = self.special_node.E

        # # Uniform local alignment fragment length distribution
        # t = special_trans
        # t.BM = log(2) - log(M) - log(M + 1)
        # t.ME = 0.0
        # for node in self.core_nodes():
        #     self.set_transition(B, node.M, t.BM)
        #     self.set_transition(node.M, E, t.ME)

        # for node in self.core_nodes()[1:]:
        #     self.set_transition(node.D, E, 0.0)

    def set_entry_transitions(
        self, entry_distr: EntryDistr, core_trans: List[Transitions]
    ):
        if entry_distr == EntryDistr.UNIFORM:
            M = self.core_length
            costs = [log(2.0 / (M * (M + 1)))] * M

        elif entry_distr == EntryDistr.OCCUPANCY:
            log_occ, logZ = _calculate_occupancy(core_trans)
            costs = [locc - logZ for locc in log_occ]

        B = self.special_node.B
        for cost, node in zip(costs, self.core_nodes()):
            self.set_transition(B, node.M, cost)

    def set_exit_transitions(self):
        E = self.special_node.E

        for node in self.core_nodes():
            self.set_transition(node.M, E, 0.0)

        for node in self.core_nodes()[1:]:
            self.set_transition(node.D, E, 0.0)

    def set_special_transitions(
        self, special_trans: SpecialTransitions, hmmer3_compat=False
    ):
        t = special_trans
        node = self.special_node
        if self._dp is not None:
            if hmmer3_compat:
                t.NN = 0.0
                t.CC = 0.0
                t.JJ = 0.0

            def dp_set(a, b, lprob):
                imm.lib.imm_dp_change_trans(
                    self._dp.imm_dp, self._hmm.imm_hmm, a.imm_state, b.imm_state, lprob,
                )

            dp_set(node.S, node.B, t.NB)
            dp_set(node.S, node.N, t.NN)
            dp_set(node.N, node.N, t.NN)
            dp_set(node.N, node.B, t.NB)

            dp_set(node.E, node.T, t.EC + t.CT)
            dp_set(node.E, node.C, t.EC + t.CC)
            dp_set(node.C, node.C, t.CC)
            dp_set(node.C, node.T, t.CT)

            dp_set(node.E, node.B, t.EJ + t.JB)
            dp_set(node.E, node.J, t.EJ + t.JJ)
            dp_set(node.J, node.J, t.JJ)
            dp_set(node.J, node.B, t.JB)

            dp_set(node.B, self._core_nodes[1].D, lprob_zero())
            dp_set(node.B, self._core_nodes[0].I, lprob_zero())
        else:
            self._dp = None

            if hmmer3_compat:
                t.NN = 0.0
                t.CC = 0.0
                t.JJ = 0.0

            self.set_transition(node.S, node.B, t.NB)
            self.set_transition(node.S, node.N, t.NN)
            self.set_transition(node.N, node.N, t.NN)
            self.set_transition(node.N, node.B, t.NB)

            self.set_transition(node.E, node.T, t.EC + t.CT)
            self.set_transition(node.E, node.C, t.EC + t.CC)
            self.set_transition(node.C, node.C, t.CC)
            self.set_transition(node.C, node.T, t.CT)

            self.set_transition(node.E, node.B, t.EJ + t.JB)
            self.set_transition(node.E, node.J, t.EJ + t.JJ)
            self.set_transition(node.J, node.J, t.JJ)
            self.set_transition(node.J, node.B, t.JB)

            # TODO: if fails for HMM having one core state
            self.set_transition(node.B, self._core_nodes[1].D, lprob_zero())
            self.set_transition(node.B, self._core_nodes[0].I, lprob_zero())

    def __str__(self):
        msg = f"{self._hmm}\n"
        msg += f"DP: {self._dp}"
        return msg


def _calculate_occupancy(core_trans: List[Transitions]) -> Tuple[List[float], float]:
    log_occ = [lprob_add(core_trans[0].MI, core_trans[0].MM)]
    for trans in core_trans[1:-1]:
        val0 = log_occ[-1] + lprob_add(trans.MM, trans.MI)
        val1 = _log1_p(log_occ[-1]) + trans.DM
        log_occ.append(lprob_add(val0, val1))

    logZ = lprob_zero()
    for i, locc in enumerate(log_occ):
        logZ = lprob_add(logZ, locc + log(len(log_occ) - i))

    return log_occ, logZ


def _log1_p(log_p: float):
    """
    Computes log(1 - p) given log(p).
    """
    from math import exp, log1p

    return log1p(-exp(log_p))



from typing import TypeVar, Union

from imm import MuteState, NormalState, Path, Results, State, Step
from nmm import AminoAlphabet, CodonState, DNAAlphabet, IUPACAminoAlphabet, RNAAlphabet

from .fragment import Fragment

AminoStep = Step[Union[NormalState, MuteState]]
AminoPath = Path[AminoStep]
AminoFragment = Fragment[AminoAlphabet, NormalState]

CodonStep = Step[Union[CodonState, MuteState]]
CodonPath = Path[CodonStep]

TState = TypeVar("TState", bound=State)

MutableState = Union[TState, MuteState]
MutableStep = Step[Union[TState, MuteState]]
MutablePath = Path[Step[Union[TState, MuteState]]]
MutableResults = Results[Union[TState, MuteState]]

HMMERAlphabet = Union[RNAAlphabet, DNAAlphabet, IUPACAminoAlphabet]


__all__ = [
    "AminoFragment",
    "AminoPath",
    "AminoStep",
    "CodonPath",
    "CodonStep",
    "HMMERAlphabet",
    "MutablePath",
    "MutableResults",
    "MutableState",
    "MutableStep",
]



from math import log
from typing import List, Mapping, NamedTuple

import hmmer_reader
from imm import lprob_zero
from nmm import IUPACAminoAlphabet

from .alphabet import infer_alphabet
from .model import Transitions
from .typing import HMMERAlphabet

__all__ = ["HMMERModel", "ModelID"]

ModelID = NamedTuple("ModelID", [("name", str), ("acc", str)])


class HMMERModel:
    """
    HMMER model.

    Parameters
    ----------
    hmmer_model
        HMMER model.
    """

    def __init__(self, hmmer_model: hmmer_reader.HMMERModel):
        self._original_symbols: str = hmmer_model.alphabet
        alphabet = infer_alphabet(self._original_symbols.encode())

        if alphabet is None:
            raise ValueError("Could not infer alphabet from HMMER model.")
        self._alphabet = alphabet

        if isinstance(self._alphabet, IUPACAminoAlphabet):
            self._null_lprobs = _null_amino_lprobs(self._original_symbols)
        else:
            k = alphabet.length
            self._null_lprobs = [log(1 / k)] * k

        self._match_lprobs = [
            self._sort(hmmer_model.match(i)) for i in range(1, hmmer_model.M + 1)
        ]
        self._insert_lprobs = [
            self._sort(hmmer_model.insert(i)) for i in range(1, hmmer_model.M + 1)
        ]
        self._model_length = hmmer_model.M

        self._transitions: List[Transitions] = []
        for m in range(0, hmmer_model.M + 1):
            t = Transitions(**hmmer_model.trans(m))
            self._transitions.append(t)

        mt = dict(hmmer_model.metadata)
        self._model_id = ModelID(mt.get("NAME", "-"), mt.get("ACC", "-"))

    @property
    def model_id(self) -> ModelID:
        return self._model_id

    @property
    def transitions(self) -> List[Transitions]:
        return self._transitions

    @property
    def model_length(self) -> int:
        return self._model_length

    @property
    def alphabet(self) -> HMMERAlphabet:
        return self._alphabet

    @property
    def null_lprobs(self) -> List[float]:
        return self._null_lprobs

    def match_lprobs(self, m: int) -> List[float]:
        return self._match_lprobs[m - 1]

    def insert_lprobs(self, m: int) -> List[float]:
        return self._insert_lprobs[m - 1]

    def _sort(self, lprobs: Mapping[str, float]) -> List[float]:
        symbols = self._alphabet.symbols.decode()
        return [lprobs.get(sym, lprob_zero()) for sym in symbols]


def _null_amino_lprobs(symbols: str):
    """
    Copy/paste from HMMER3 amino acid frequences infered form Swiss-Prot 50.8,
    (Oct 2006), counting over 85956127 (86.0M) residues.
    """
    lprobs = {
        "A": log(0.0787945),
        "C": log(0.0151600),
        "D": log(0.0535222),
        "E": log(0.0668298),
        "F": log(0.0397062),
        "G": log(0.0695071),
        "H": log(0.0229198),
        "I": log(0.0590092),
        "K": log(0.0594422),
        "L": log(0.0963728),
        "M": log(0.0237718),
        "N": log(0.0414386),
        "P": log(0.0482904),
        "Q": log(0.0395639),
        "R": log(0.0540978),
        "S": log(0.0683364),
        "T": log(0.0540687),
        "V": log(0.0673417),
        "W": log(0.0114135),
        "Y": log(0.0304133),
    }
    return [lprobs.get(sym, lprob_zero()) for sym in list(symbols)]



import os
import re
from collections import OrderedDict
from typing import IO, Dict, List, Set, Tuple

import click
from fasta_reader import FASTAWriter, open_fasta
from hmmer import HMMER
from hmmer.typing import DomTBLRow
from hmmer_reader import num_models, open_hmmer
from nmm import AminoAlphabet, BaseAlphabet, IUPACAminoAlphabet
from tqdm import tqdm

from iseq.alphabet import alphabet_name, infer_fasta_alphabet, infer_hmmer_alphabet
from iseq.codon_table import CodonTable
from iseq.gff import read as read_gff
from iseq.hmmer_model import HMMERModel
from iseq.protein import create_profile2

from .debug_writer import DebugWriter
from .output_writer import OutputWriter


@click.command()
@click.argument(
    "profile",
    type=click.Path(exists=True, dir_okay=False, readable=True, resolve_path=False),
)
@click.argument("target", type=click.File("r"))
@click.option(
    "--epsilon", type=float, default=1e-2, help="Indel probability. Defaults to 1e-2."
)
@click.option(
    "--output",
    type=click.Path(exists=False, dir_okay=False, writable=True, resolve_path=True),
    help="Save results to OUTPUT (GFF format).",
    default="output.gff",
)
@click.option(
    "--ocodon",
    type=click.Path(exists=False, dir_okay=False, writable=True, resolve_path=True),
    help="Save codon sequences to OCODON (FASTA format).",
    default="ocodon.fasta",
)
@click.option(
    "--oamino",
    type=click.Path(exists=False, dir_okay=False, writable=True, resolve_path=True),
    help="Save amino acid sequences to OAMINO (FASTA format).",
    default="oamino.fasta",
)
@click.option(
    "--quiet/--no-quiet", "-q/-nq", help="Disable standard output.", default=False,
)
@click.option(
    "--window",
    type=int,
    help="Window length. Defaults to zero, which means no window.",
    default=0,
)
@click.option(
    "--odebug",
    type=click.File("w"),
    help="Save debug info into a tab-separated values file.",
    default=os.devnull,
)
@click.option(
    "--max-e-value",
    type=float,
    default=10.0,
    help="Filter out items for which E-value > --max-e-evalue.",
)
@click.option(
    "--hit-prefix", help="Hit prefix. Defaults to `item`.", default="item", type=str,
)
@click.option(
    "--cut-ga/--no-cut-ga",
    help="Enable use of profile's GA gathering cutoffs to set all thresholding. Defaults to True.",
    default=True,
)
def pscan2(
    profile,
    target,
    epsilon: float,
    output,
    ocodon,
    oamino,
    quiet,
    window: int,
    odebug,
    max_e_value: float,
    hit_prefix: str,
    cut_ga: bool,
):
    """
    Search nucleotide sequence(s) against a protein profiles database.

    An OUTPUT line determines an association between a TARGET subsequence and
    a PROFILE protein profile. An association maps a target subsequence to a
    profile and represents a potential homology. Expect many false positive
    associations as we are not filtering out by statistical significance.
    """

    owriter = OutputWriter(output, item_prefix=hit_prefix)
    cwriter = FASTAWriter(ocodon)
    awriter = FASTAWriter(oamino)
    dwriter = DebugWriter(odebug)

    with open(profile, "r") as file:
        profile_abc = infer_profile_alphabet(file)
    target_abc = infer_target_alphabet(target)

    assert isinstance(target_abc, BaseAlphabet) and isinstance(
        profile_abc, AminoAlphabet
    )

    gcode = CodonTable(target_abc, IUPACAminoAlphabet())

    with open_fasta(target) as fasta:
        targets = list(fasta)

    total = num_models(profile)
    for plain_model in tqdm(
        open_hmmer(profile), desc="Models", total=total, disable=quiet
    ):
        hmodel = HMMERModel(plain_model)
        prof = create_profile2(hmodel, gcode.base_alphabet, window, epsilon)

        for tgt in tqdm(targets, desc="Targets", leave=False, disable=quiet):
            seq = prof.create_sequence(tgt.sequence.encode())
            search_results = prof.search(seq)
            ifragments = search_results.ifragments()
            seqid = f"{tgt.id}"

            for ifrag in ifragments:
                start = ifrag.interval.start
                stop = ifrag.interval.stop
                item_id = owriter.write_item(
                    seqid,
                    alphabet_name(seq.alphabet),
                    prof.profid,
                    alphabet_name(prof.alphabet),
                    start,
                    stop,
                    prof.window_length,
                    {"Epsilon": epsilon},
                )
                codon_frag = ifrag.fragment.decode()
                cwriter.write_item(item_id, str(codon_frag.sequence))
                amino_frag = codon_frag.decode(gcode)
                awriter.write_item(item_id, str(amino_frag.sequence))

            if odebug is not os.devnull:
                for i in search_results.debug_table():
                    dwriter.write_row(seqid, i)

    owriter.close()
    cwriter.close()
    awriter.close()
    odebug.close_intelligently()

    if not quiet:
        click.echo("Computing e-values... ", nl=False)
    hmmer = HMMER(profile)
    if not hmmer.is_pressed:
        hmmer.press()
    result = hmmer.scan(oamino, "/dev/null", domtblout=True, cut_ga=cut_ga)
    score_table = ScoreTable(result.domtbl)

    update_gff_file(output, score_table, max_e_value)
    target_set = target_set_from_gff_file(output)
    update_fasta_file(ocodon, target_set)
    update_fasta_file(oamino, target_set)
    if not quiet:
        click.echo("done.")

    def rep(k: str):
        return int(k.replace(hit_prefix, ""))

    if not quiet:
        click.echo("Translating files... ", nl=False)
    items = list(sorted(list(target_set), key=rep))
    target_map = {item: f"{hit_prefix}{i+1}" for i, item in enumerate(items)}
    translate_gff_file(output, target_map)
    translate_fasta_file(ocodon, target_map)
    translate_fasta_file(oamino, target_map)
    if not quiet:
        click.echo("done.")


def infer_profile_alphabet(profile: IO[str]):
    hmmer = open_hmmer(profile)
    hmmer_alphabet = infer_hmmer_alphabet(hmmer)
    profile.seek(0)
    if hmmer_alphabet is None:
        raise click.UsageError("Could not infer alphabet from PROFILE.")
    return hmmer_alphabet


def infer_target_alphabet(target: IO[str]):
    fasta = open_fasta(target)
    target_alphabet = infer_fasta_alphabet(fasta)
    target.seek(0)
    if target_alphabet is None:
        raise click.UsageError("Could not infer alphabet from TARGET.")
    return target_alphabet


class ScoreTable:
    def __init__(self, domtbldata: List[DomTBLRow]):
        self._tbldata: Dict[Tuple[str, str, str], DomTBLRow] = {}
        for line in domtbldata:
            key = (line.query.name, line.target.name, line.target.accession)
            if key in self._tbldata:
                prev_score = float(self._tbldata[key].full_sequence.score)
                if float(line.full_sequence.score) > prev_score:
                    self._tbldata[key] = line
            else:
                self._tbldata[key] = line

    def e_value(self, target_id: str, profile_name: str, profile_acc: str) -> str:
        key = (target_id, profile_name, profile_acc)
        return self._tbldata[key].full_sequence.e_value

    def has(self, target_id: str, profile_name: str, profile_acc: str) -> bool:
        key = (target_id, profile_name, profile_acc)
        return key in self._tbldata


def update_gff_file(filepath, score_table: ScoreTable, max_e_value: float):
    import in_place

    with in_place.InPlace(filepath) as file:
        for row in file:
            row = row.rstrip()
            if row.startswith("#"):
                file.write(row)
                file.write("\n")
                continue

            match = re.match(r"^(.+\t)([^\t]+)$", row)
            if match is None:
                file.write(row)
                file.write("\n")
                continue

            left = match.group(1)
            right = match.group(2)

            if right == ".":
                file.write(row)
                file.write("\n")
                continue

            fields_list = []
            for v in right.split(";"):
                name, value = v.split("=", 1)
                fields_list.append((name, value))

            attr = OrderedDict(fields_list)

            key = (attr["ID"], attr["Profile_name"], attr["Profile_acc"])
            if not score_table.has(*key):
                continue

            attr["E-value"] = score_table.e_value(*key)
            if float(attr["E-value"]) > max_e_value:
                continue

            file.write(left + ";".join(k + "=" + v for k, v in attr.items()))
            file.write("\n")


def translate_gff_file(filepath, target_map: Dict[str, str]):
    import in_place

    with in_place.InPlace(filepath) as file:
        for row in file:
            row = row.rstrip()
            if row.startswith("#"):
                file.write(row)
                file.write("\n")
                continue

            match = re.match(r"^(.+\t)([^\t]+)$", row)
            if match is None:
                file.write(row)
                file.write("\n")
                continue

            left = match.group(1)
            right = match.group(2)

            if right == ".":
                file.write(row)
                file.write("\n")
                continue

            fields_list = []
            for v in right.split(";"):
                name, value = v.split("=", 1)
                fields_list.append((name, value))

            attr = OrderedDict(fields_list)

            attr["ID"] = target_map[attr["ID"]]
            file.write(left + ";".join(k + "=" + v for k, v in attr.items()))
            file.write("\n")


def target_set_from_gff_file(filepath) -> Set[str]:
    gff = read_gff(filepath)
    df = gff.to_dataframe()
    if len(df) == 0:
        return set()
    return set(df["att_ID"].tolist())


def update_fasta_file(filepath, target_set: Set[str]):
    with open_fasta(filepath) as fasta:
        targets = list(fasta)

    with FASTAWriter(filepath) as writer:
        for target in targets:
            tgt_id = target.id
            if tgt_id in target_set:
                writer.write_item(target.defline, target.sequence)


def translate_fasta_file(filepath, target_map: Dict[str, str]):
    with open_fasta(filepath) as fasta:
        targets = list(fasta)

    with FASTAWriter(filepath) as writer:
        for target in targets:
            old_tgt_id = target.id
            tgt_id = target_map[old_tgt_id]
            if target.has_desc:
                defline = tgt_id + " " + target.desc
            else:
                defline = tgt_id
            writer.write_item(defline, target.sequence)



from nmm import Codon, DNAAlphabet, IUPACAminoAlphabet, RNAAlphabet

from iseq.codon_table import CodonTable
from iseq.gencode import GeneticCode


def test_codon_table_dna():
    base_abc = DNAAlphabet()
    amino_abc = IUPACAminoAlphabet()

    table = CodonTable(base_abc, amino_abc)

    assert len(table.codons(b"P")) == 4

    assert Codon.create(b"CCT", base_abc) in table.codons(b"P")
    assert Codon.create(b"CCC", base_abc) in table.codons(b"P")
    assert Codon.create(b"CCA", base_abc) in table.codons(b"P")
    assert Codon.create(b"CCG", base_abc) in table.codons(b"P")

    assert len(table.codons(b"W")) == 1
    assert Codon.create(b"TGG", base_abc) in table.codons(b"W")

    assert table.amino_acid(Codon.create(b"ATG", base_abc)) == b"M"
    assert len(table.amino_acids) == 20
    assert b"R" in table.amino_acids

    assert len(table.stop_codons) == 3
    assert Codon.create(b"TAA", base_abc) in table.stop_codons
    assert Codon.create(b"TAG", base_abc) in table.stop_codons
    assert Codon.create(b"TGA", base_abc) in table.stop_codons

    assert len(set(table.codons())) == 64


def test_codon_table_dna_id33():
    base_abc = DNAAlphabet()
    amino_abc = IUPACAminoAlphabet()

    gencode = GeneticCode(id=33)
    table = CodonTable(base_abc, amino_abc, gencode)

    assert len(table.codons(b"P")) == 4

    assert Codon.create(b"CCT", base_abc) in table.codons(b"P")
    assert Codon.create(b"CCC", base_abc) in table.codons(b"P")
    assert Codon.create(b"CCA", base_abc) in table.codons(b"P")
    assert Codon.create(b"CCG", base_abc) in table.codons(b"P")

    assert len(table.codons(b"W")) == 2
    assert Codon.create(b"TGG", base_abc) in table.codons(b"W")
    assert Codon.create(b"TGA", base_abc) in table.codons(b"W")

    assert len(table.codons(b"T")) == 4
    assert Codon.create(b"ACT", base_abc) in table.codons(b"T")
    assert Codon.create(b"ACC", base_abc) in table.codons(b"T")
    assert Codon.create(b"ACA", base_abc) in table.codons(b"T")
    assert Codon.create(b"ACG", base_abc) in table.codons(b"T")

    assert table.amino_acid(Codon.create(b"ATG", base_abc)) == b"M"
    assert len(table.amino_acids) == 20
    assert b"R" in table.amino_acids

    assert len(table.start_codons) == 4
    assert Codon.create(b"TTG", base_abc) in table.start_codons
    assert Codon.create(b"CTG", base_abc) in table.start_codons
    assert Codon.create(b"ATG", base_abc) in table.start_codons
    assert Codon.create(b"GTG", base_abc) in table.start_codons

    assert len(table.stop_codons) == 1
    assert Codon.create(b"TAG", base_abc) in table.stop_codons

    assert len(set(table.codons())) == 64


def test_codon_table_rna():
    base_abc = RNAAlphabet()
    amino_abc = IUPACAminoAlphabet()

    table = CodonTable(base_abc, amino_abc)

    assert len(table.codons(b"P")) == 4

    assert Codon.create(b"CCU", base_abc) in table.codons(b"P")
    assert Codon.create(b"CCC", base_abc) in table.codons(b"P")
    assert Codon.create(b"CCA", base_abc) in table.codons(b"P")
    assert Codon.create(b"CCG", base_abc) in table.codons(b"P")

    assert table.amino_acid(Codon.create(b"AUG", base_abc)) == b"M"
    assert len(table.amino_acids) == 20
    assert b"R" in table.amino_acids

    assert len(table.start_codons) == 3
    assert Codon.create(b"UUG", base_abc) in table.start_codons
    assert Codon.create(b"CUG", base_abc) in table.start_codons
    assert Codon.create(b"AUG", base_abc) in table.start_codons

    assert len(table.stop_codons) == 3
    assert Codon.create(b"UAA", base_abc) in table.stop_codons
    assert Codon.create(b"UAG", base_abc) in table.stop_codons
    assert Codon.create(b"UGA", base_abc) in table.stop_codons

    assert len(set(table.codons())) == 64



from typing import IO

from iseq.result import DebugRow


class DebugWriter:
    def __init__(self, file: IO[str]):

        self._file = file
        self._file.write(
            "defline\twindow\tstart\tstop\talt_viterbi_score\tnull_viterbi_score\n"
        )

    def write_row(
        self, defline: str, debug_row: DebugRow,
    ):
        """
        Write item.
        """
        wnum = debug_row.window_number
        start = debug_row.window.start + 1
        stop = debug_row.window.stop
        alt_vit = debug_row.alt_viterbi_score
        null_vit = debug_row.null_viterbi_score
        self._file.write(f"{defline}\t{wnum}\t{start}\t{stop}\t{alt_vit}\t{null_vit}\n")

    def close(self):
        """
        Close the associated stream.
        """
        self._file.close()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        del exception_type
        del exception_value
        del traceback
        self.close()



import os
from math import isfinite
from pathlib import Path

import pytest
from fasta_reader import open_fasta
from hmmer_reader import open_hmmer
from imm import Sequence
from imm.testing import assert_allclose

from iseq.example import example_filepath
from iseq.hmmer3 import create_profile
from iseq.hmmer_model import HMMERModel


@pytest.mark.slow
def test_hmmer3_pfam_viterbi_scores_compat(tmp_path):
    os.chdir(tmp_path)
    db_filepath = example_filepath("Pfam-A.hmm")
    target_filepath = example_filepath("A0ALD9.fasta")
    iseq_scores = loadtxt(example_filepath("Pfam-A_iseq_viterbi_scores.txt"))

    with open_fasta(target_filepath) as fasta:
        target = list(fasta)[0]

    actual_scores = []
    for hmmprof in open_hmmer(db_filepath):
        prof = create_profile(HMMERModel(hmmprof), hmmer3_compat=True)
        seq = Sequence.create(target.sequence.encode(), prof.alphabet)
        search_results = prof.search(seq)
        score = search_results.results[0].alt_viterbi_score
        actual_scores.append(score)

    iseq_scores = loadtxt(example_filepath("Pfam-A_iseq_viterbi_scores.txt"))
    assert_allclose(actual_scores, iseq_scores)

    hmmer3_scores = loadtxt(example_filepath("Pfam-A_hmmer3.3_viterbi_scores.txt"))
    ok = [i for i, score in enumerate(hmmer3_scores) if isfinite(score)]

    actual_scores = [actual_scores[i] for i in ok]
    hmmer3_scores = [hmmer3_scores[i] for i in ok]

    assert_allclose(actual_scores, hmmer3_scores, 3e-2)


def test_hmmer3_viterbi_dna_scores_compat():
    hmmfile = example_filepath("2OG-FeII_Oxy_3-nt.hmm")
    hmmprof = open_hmmer(hmmfile).read_model()
    prof = create_profile(HMMERModel(hmmprof), hmmer3_compat=True)
    for align in ["local", "unilocal", "glocal", "uniglocal"]:

        fastafile = f"2OG-FeII_Oxy_3-nt_{align}.fasta"
        hmmer3_vitfile = f"hmmer3_2OG-FeII_Oxy_3-nt_{align}.fasta.viterbi"
        iseq_vitfile = f"iseq_2OG-FeII_Oxy_3-nt_{align}.fasta.viterbi"

        hmmer3_scores = loadtxt(example_filepath(hmmer3_vitfile))
        iseq_scores = loadtxt(example_filepath(iseq_vitfile))

        # HMMER3 viterbi filter has very low accuracy (2 bytes of integer arithmetic)
        # while we use 8 bytes of floating point arithmetic. Therefore we have
        # to allow for relatively high viterbi score differences.
        ok = [i for i, s in enumerate(hmmer3_scores) if isfinite(s) and abs(s) > 1]

        hmmer3_scores = [hmmer3_scores[i] for i in ok]
        iseq_scores = [iseq_scores[i] for i in ok]

        assert_allclose(iseq_scores, hmmer3_scores, rtol=7e-2)

        actual_scores = []
        with open_fasta(example_filepath(fastafile)) as fasta:
            for target in fasta:
                seq = Sequence.create(target.sequence.encode(), prof.alphabet)
                search_results = prof.search(seq)
                score = search_results.results[0].alt_viterbi_score
                actual_scores.append(score)

        iseq_scores = loadtxt(example_filepath(iseq_vitfile))
        assert_allclose(actual_scores, iseq_scores)


def test_hmmer3_viterbi_amino_scores_compat():
    hmmfile = example_filepath("2OG-FeII_Oxy_3.hmm")
    hmmprof = open_hmmer(hmmfile).read_model()
    prof = create_profile(HMMERModel(hmmprof), hmmer3_compat=True)
    for align in ["local", "unilocal", "glocal", "uniglocal"]:

        fastafile = f"2OG-FeII_Oxy_3_{align}.fasta"
        hmmer3_vitfile = f"hmmer3_2OG-FeII_Oxy_3_{align}.fasta.viterbi"
        iseq_vitfile = f"iseq_2OG-FeII_Oxy_3_{align}.fasta.viterbi"

        hmmer3_scores = loadtxt(example_filepath(hmmer3_vitfile))
        iseq_scores = loadtxt(example_filepath(iseq_vitfile))

        # HMMER3 viterbi filter has very low accuracy (2 bytes of integer arithmetic)
        # while we use 8 bytes of floating point arithmetic. Therefore we have
        # to allow for relatively high viterbi score differences.
        ok = [i for i, s in enumerate(hmmer3_scores) if isfinite(s)]

        hmmer3_scores = [hmmer3_scores[i] for i in ok]
        iseq_scores = [iseq_scores[i] for i in ok]

        assert_allclose(iseq_scores, hmmer3_scores, rtol=1e-2)

        actual_scores = []
        with open_fasta(example_filepath(fastafile)) as fasta:
            for target in fasta:
                seq = Sequence.create(target.sequence.encode(), prof.alphabet)
                search_results = prof.search(seq)
                score = search_results.results[0].alt_viterbi_score
                actual_scores.append(score)

        iseq_scores = loadtxt(example_filepath(iseq_vitfile))
        assert_allclose(actual_scores, iseq_scores)


def loadtxt(filepath: Path):
    arr = []
    with open(filepath, "r") as file:
        for line in file:
            arr.append(float(line))
    return arr



import click

from iseq.gff import read as read_gff


@click.command()
@click.argument("gff_file", type=click.File("r"))
@click.option(
    "--output",
    type=click.Path(exists=False, dir_okay=False, writable=True, resolve_path=True),
    help="Save figure to OUTPUT.",
    default=None,
)
@click.option("--width", help="Figure width.", default=None, type=float)
@click.option("--height", help="Figure height.", default=None, type=float)
@click.option(
    "--quiet/--no-quiet", "-q/-nq", help="Disable verbosity.", default=False,
)
def profile_dist(gff_file, output, width, height, quiet: bool):
    """
    Plot profiles distribution.
    """
    import seaborn as sns
    from matplotlib import pyplot as plt

    gff = read_gff(gff_file, verbose=not quiet)
    df = gff.to_dataframe()

    sns.set(color_codes=True)
    figsize = list(plt.rcParams.get("figure.figsize"))
    if width is not None:
        figsize[0] = float(width)
    if height is not None:
        figsize[1] = float(height)

    fig, ax = plt.subplots(figsize=figsize)
    ax.set_title("Distribution of Profiles")
    df["Profile"] = df["att_Profile_name"]
    prof_size = df.groupby("Profile").size()
    prof_size.plot.bar(axes=ax)
    ax.set_ylabel("Number of subsequences")
    fig.tight_layout()

    if not quiet:
        click.echo("Plotting... ", nl=False)
    if output is None:
        plt.show()
    else:
        plt.savefig(output)

    if not quiet:
        click.echo("done.", nl=True)
        if output is not None:
            click.echo(f"Saved to {output}.")



import click

from .e_value_dist import e_value_dist
from .length_ratio import length_ratio
from .profile_dist import profile_dist


@click.group(name="plot", context_settings=dict(help_option_names=["-h", "--help"]))
def plot():
    """
    Plot.
    """


plot.add_command(e_value_dist)
plot.add_command(length_ratio)
plot.add_command(profile_dist)



from . import typing
from ._fragment import ProteinFragment
from ._profile import ProteinProfile, create_profile, create_profile2

__all__ = [
    "ProteinFragment",
    "ProteinProfile",
    "create_profile",
    "typing",
    "create_profile2",
]



def pytest_addoption(parser):
    try:
        parser.addoption(
            "--runslow", action="store_true", default=False, help="run slow tests",
        )
    except ValueError as e:
        if "already added" not in e.args[0]:
            raise e


def pytest_collection_modifyitems(config, items):
    import pytest

    if config.getoption("--runslow"):
        # --runslow given in cli: do not skip slow tests
        return

    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)



from math import log
from typing import List

from imm import MuteState, Path, Sequence, SequenceABC
from nmm import BaseAlphabet, Codon, CodonProb, CodonState, FrameState

from iseq.codon import CodonFragment
from iseq.fragment import Fragment
from iseq.typing import CodonPath, CodonStep

from .typing import ProteinStep

__all__ = ["ProteinFragment"]


class ProteinFragment(Fragment[BaseAlphabet, FrameState]):
    def __init__(
        self,
        sequence: SequenceABC[BaseAlphabet],
        path: Path[ProteinStep],
        homologous: bool,
    ):
        super().__init__(sequence, path, homologous)

    def decode(self) -> CodonFragment:
        codons: List[Codon] = []
        steps: List[CodonStep] = []

        start = 0
        for step in self.path:
            if isinstance(step.state, MuteState):

                mstate = MuteState.create(step.state.name, step.state.alphabet)
                steps.append(CodonStep.create(mstate, 0))

            elif isinstance(step.state, FrameState):

                subseq = self.sequence[start : start + step.seq_len]
                codon = step.state.decode(subseq)[1]
                codons.append(codon)

                name = step.state.name
                codonp = CodonProb.create(self.sequence.alphabet)
                codonp.set_lprob(codon, log(1.0))
                cstate = CodonState.create(name, codonp)
                steps.append(CodonStep.create(cstate, 3))

            else:
                raise RuntimeError("Wrong state type.")

            start += step.seq_len

        FrameSequence = Sequence[BaseAlphabet]
        abc = self.sequence.alphabet
        sequence = FrameSequence.create(b"".join(c.symbols for c in codons), abc)
        return CodonFragment(sequence, CodonPath.create(steps), self.homologous)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}:{str(self)}>"



from math import log

from hmmer_reader import open_hmmer
from imm.testing import assert_allclose
from nmm import DNAAlphabet, IUPACAminoAlphabet

from iseq.example import example_filepath
from iseq.hmmer_model import HMMERModel


def test_hmmdata_amino():
    filepath = example_filepath("PF03373.hmm")
    with open_hmmer(filepath) as reader:
        hmm_data = HMMERModel(reader.read_model())

    assert isinstance(hmm_data.alphabet, IUPACAminoAlphabet)
    assert_allclose(hmm_data.null_lprobs[3], -2.7056061901315998)
    assert_allclose(hmm_data.match_lprobs(6)[3], -3.36585)
    assert_allclose(hmm_data.insert_lprobs(1)[0], -2.68618)


def test_hmmdata_dna():
    filepath = example_filepath("ecori.hmm")
    with open_hmmer(filepath) as reader:
        hmm_data = HMMERModel(reader.read_model())

    assert isinstance(hmm_data.alphabet, DNAAlphabet)
    assert_allclose(hmm_data.null_lprobs[3], log(1 / 4))
    assert_allclose(hmm_data.match_lprobs(2)[2], -2.42319)
    assert_allclose(hmm_data.insert_lprobs(1)[0], -1.38629)



import warnings
from typing import Dict, List, Optional, Union

from nmm import AminoAlphabet, Codon, DNAAlphabet, RNAAlphabet

from .gencode import GeneticCode

with warnings.catch_warnings():
    warnings.simplefilter("ignore", category=PendingDeprecationWarning)
    from Bio.Data.CodonTable import (
        NCBICodonTable,
        unambiguous_dna_by_id,
        unambiguous_dna_by_name,
    )

__all__ = ["CodonTable"]


class CodonTable:
    """
    Codon table.

    Parameters
    ----------
    base_abc
        Base alphabet.
    amino_abc
        Amino acid alphabet.
    gencode
        NCBI `genetic code`_. Defaults to `GeneticCode("Standard")`.

    .. _genetic code: https://www.ncbi.nlm.nih.gov/Taxonomy/Utils/wprintgc.cgi
    """

    def __init__(
        self,
        base_abc: Union[DNAAlphabet, RNAAlphabet],
        amino_abc: AminoAlphabet,
        gencode: Optional[GeneticCode] = None,
    ):

        self._base_alphabet = base_abc
        self._amino_alphabet = amino_abc

        if gencode is None:
            gencode = GeneticCode("Standard")
        table = translation_table(gencode)

        def replace(seq: bytes):
            if isinstance(base_abc, RNAAlphabet):
                seq = seq.replace(b"T", b"U")
            return seq

        self._codons: Dict[bytes, List[Codon]] = {}

        self._start_codons = []
        for t, a in table.forward_table.items():
            triplet = replace(t.encode())
            aa = a.encode()
            if aa not in self._codons:
                self._codons[aa] = []

            codon = Codon.create(triplet, base_abc)
            if t in table.start_codons:
                self._start_codons.append(codon)
            self._codons[aa].append(codon)

        self._amino_acid: Dict[Codon, bytes] = {}
        for aa, codons in self._codons.items():
            for codon in codons:
                self._amino_acid[codon] = aa

        self._stop_codons = []
        for t in table.stop_codons:
            triplet = replace(t.encode())

            self._stop_codons.append(Codon.create(triplet, base_abc))

        assert len(self._amino_acid) <= 64
        assert len(self._amino_acid) + len(self._stop_codons) == 64

    @property
    def start_codons(self) -> List[Codon]:
        return self._start_codons

    @property
    def stop_codons(self) -> List[Codon]:
        return self._stop_codons

    def codons(self, amino_acid: Optional[bytes] = None) -> List[Codon]:

        if amino_acid is None:
            all_codons: List[Codon] = []
            for aa in self.amino_acids:
                all_codons += self.codons(aa)
            all_codons += self.stop_codons
            return list(sorted(all_codons))

        amino_acid = amino_acid.upper()
        return self._codons.get(amino_acid, [])

    def amino_acid(self, codon: Codon) -> bytes:
        return self._amino_acid[codon]

    def decode(self, codon: Codon) -> bytes:
        if codon in self._amino_acid:
            return self._amino_acid[codon]
        assert codon in self._stop_codons
        return self._amino_alphabet.stop_symbol

    @property
    def amino_acids(self) -> List[bytes]:
        return list(set(self._amino_acid.values()))

    @property
    def base_alphabet(self) -> Union[DNAAlphabet, RNAAlphabet]:
        return self._base_alphabet

    @property
    def amino_alphabet(self) -> AminoAlphabet:
        return self._amino_alphabet


def translation_table(gencode: GeneticCode) -> NCBICodonTable:

    if gencode.id not in unambiguous_dna_by_id:
        names = str(list(unambiguous_dna_by_name.keys()))
        msg = f"Unknown translation table {gencode.name}. Possible names are: {names}."
        raise ValueError(msg)

    return unambiguous_dna_by_id[gencode.id]



# NCBI genetic code table version 4.6
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple

__all__ = ["GeneticCode"]

genetic_codes: List[Tuple[str, str, int]] = []
names: Dict[str, int] = {}
alt_names: Dict[str, int] = {}
ids: Dict[int, int] = {}


@dataclass
class GeneticCode:
    name: str
    alt_name: str
    id: int

    def __init__(
        self,
        name: Optional[str] = None,
        alt_name: Optional[str] = None,
        id: Optional[int] = None,
    ):
        n = sum([name is None, alt_name is None, id is None])

        if n != 2:
            raise ValueError("You must use one, and only one, parameter.")

        if name is not None:
            if name in names:
                self.name, self.alt_name, self.id = genetic_codes[names[name]]
                return
            raise ValueError(f"Unknown name {name}.")

        if alt_name is not None:
            if alt_name in alt_names:
                self.name, self.alt_name, self.id = genetic_codes[alt_names[alt_name]]
                return
            raise ValueError(f"Unknown alternative name {alt_name}.")

        assert id is not None
        self.name, self.alt_name, self.id = genetic_codes[ids[id]]


def register_ncbi_genetic_code(name: str, alt_name: str, id: int):
    names[name] = len(genetic_codes)
    if alt_name != "":
        alt_names[alt_name] = len(genetic_codes)
    ids[id] = len(genetic_codes)
    genetic_codes.append((name, alt_name, id))


register_ncbi_genetic_code(
    "Standard", "SGC0", 1,
)

register_ncbi_genetic_code(
    "Vertebrate Mitochondrial", "SGC1", 2,
)

register_ncbi_genetic_code(
    "Yeast Mitochondrial", "SGC2", 3,
)

register_ncbi_genetic_code(
    "Mold Mitochondrial; Protozoan Mitochondrial; Coelenterate "
    "Mitochondrial; Mycoplasma; Spiroplasma",
    "SGC3",
    4,
)

register_ncbi_genetic_code(
    "Invertebrate Mitochondrial", "SGC4", 5,
)

register_ncbi_genetic_code(
    "Ciliate Nuclear; Dasycladacean Nuclear; Hexamita Nuclear", "SGC5", 6,
)

register_ncbi_genetic_code(
    "Echinoderm Mitochondrial; Flatworm Mitochondrial", "SGC8", 9,
)

register_ncbi_genetic_code(
    "Euplotid Nuclear", "SGC9", 10,
)

register_ncbi_genetic_code(
    "Bacterial, Archaeal and Plant Plastid", "", 11,
)

register_ncbi_genetic_code(
    "Alternative Yeast Nuclear", "", 12,
)

register_ncbi_genetic_code(
    "Ascidian Mitochondrial", "", 13,
)

register_ncbi_genetic_code(
    "Alternative Flatworm Mitochondrial", "", 14,
)

register_ncbi_genetic_code(
    "Blepharisma Macronuclear", "", 15,
)

register_ncbi_genetic_code(
    "Chlorophycean Mitochondrial", "", 16,
)

register_ncbi_genetic_code(
    "Trematode Mitochondrial", "", 21,
)

register_ncbi_genetic_code(
    "Scenedesmus obliquus Mitochondrial", "", 22,
)

register_ncbi_genetic_code(
    "Thraustochytrium Mitochondrial", "", 23,
)

register_ncbi_genetic_code(
    "Rhabdopleuridae Mitochondrial", "", 24,
)

register_ncbi_genetic_code(
    "Candidate Division SR1 and Gracilibacteria", "", 25,
)

register_ncbi_genetic_code(
    "Pachysolen tannophilus Nuclear", "", 26,
)

register_ncbi_genetic_code(
    "Karyorelict Nuclear", "", 27,
)

register_ncbi_genetic_code(
    "Condylostoma Nuclear", "", 28,
)

register_ncbi_genetic_code(
    "Mesodinium Nuclear", "", 29,
)

register_ncbi_genetic_code(
    "Peritrich Nuclear", "", 30,
)

register_ncbi_genetic_code(
    "Blastocrithidia Nuclear", "", 31,
)

register_ncbi_genetic_code(
    "Balanophoraceae Plastid", "", 32,
)

register_ncbi_genetic_code(
    "Cephalodiscidae Mitochondrial", "", 33,
)



from __future__ import annotations

from math import log
from typing import List, Type

import nmm
from imm import (
    Interval,
    MuteState,
    Path,
    Sequence,
    SequenceABC,
    lprob_add,
    lprob_normalize,
    lprob_zero,
)
from nmm import AminoTable, BaseAlphabet, BaseTable, CodonProb, FrameState, codon_iter

from iseq import wrap
from iseq.codon_table import CodonTable
from iseq.hmmer_model import HMMERModel
from iseq.model import EntryDistr, Transitions
from iseq.profile import Profile, ProfileID

from ._fragment import ProteinFragment
from .typing import (
    ProteinAltModel,
    ProteinNode,
    ProteinNullModel,
    ProteinSearchResults,
    ProteinSpecialNode,
    ProteinStep,
)

__all__ = ["ProteinProfile", "create_profile", "create_profile2"]


class ProteinProfile(Profile[BaseAlphabet, FrameState]):
    @classmethod
    def create(
        cls: Type[ProteinProfile],
        profid: ProfileID,
        factory: ProteinStateFactory,
        null_aminot: AminoTable,
        core_nodes: List[ProteinNode],
        core_trans: List[Transitions],
        entry_distr: EntryDistr,
    ) -> ProteinProfile:

        base_alphabet = factory.genetic_code.base_alphabet

        R = factory.create(b"R", null_aminot)
        null_model = ProteinNullModel.create(R)

        special_node = ProteinSpecialNode(
            S=MuteState.create(b"S", base_alphabet),
            N=factory.create(b"N", null_aminot),
            B=MuteState.create(b"B", base_alphabet),
            E=MuteState.create(b"E", base_alphabet),
            J=factory.create(b"J", null_aminot),
            C=factory.create(b"C", null_aminot),
            T=MuteState.create(b"T", base_alphabet),
        )

        alt_model = ProteinAltModel.create(
            special_node, core_nodes, core_trans, entry_distr,
        )
        # alt_model.set_fragment_length(self._special_transitions)
        return cls(profid, base_alphabet, null_model, alt_model, False)

    @property
    def epsilon(self) -> float:
        nodes = self._alt_model.core_nodes()
        return nodes[0].M.epsilon

    @classmethod
    def create_from_binary(
        cls: Type[ProteinProfile],
        profid: ProfileID,
        null_model: nmm.Model,
        alt_model: nmm.Model,
    ):
        special_node = wrap.special_node(alt_model.hmm)
        core_nodes = wrap.core_nodes(alt_model.hmm)
        alt = ProteinAltModel.create_from_hmm(
            special_node, core_nodes, alt_model.hmm, alt_model.dp
        )
        null = ProteinNullModel.create_from_hmm(null_model.hmm)
        return cls(profid, alt_model.hmm.alphabet, null, alt, False)

    @property
    def window_length(self) -> int:
        return super().window_length

    @window_length.setter
    def window_length(self, length: int) -> None:
        if length < -1:
            raise ValueError("Length must be greater than or equal to -1.")

        if length == -1:
            length = 2 * 3 * self._alt_model.core_length
        self._window_length = length

    def create_sequence(self, sequence: bytes) -> Sequence:
        return Sequence.create(sequence, self.alphabet)

    @property
    def null_model(self) -> ProteinNullModel:
        return self._null_model

    @property
    def alt_model(self) -> ProteinAltModel:
        return self._alt_model

    def search(self, sequence: SequenceABC[BaseAlphabet]) -> ProteinSearchResults:

        # special_trans = self._get_target_length_model(len(sequence))
        # self._alt_model.set_special_transitions(special_trans)
        # self._null_model.set_special_transitions(special_trans)

        # alt_results = self.alt_model.viterbi(sequence, window_length)
        self._set_target_length_model(len(sequence))
        alt_results = self._alt_model.viterbi(sequence, self.window_length)

        def create_fragment(
            seq: SequenceABC[BaseAlphabet], path: Path[ProteinStep], homologous: bool
        ):
            return ProteinFragment(seq, path, homologous)

        search_results = ProteinSearchResults(sequence, create_fragment)

        for alt_result in alt_results:
            subseq = alt_result.sequence
            # TODO: temporary fix for reading from binary file
            # and consequently alt and null model having different alphabets
            s = Sequence.create(bytes(subseq), self._null_model.hmm.alphabet)
            viterbi_score0 = self._null_model.likelihood(s)
            # viterbi_score0 = self._null_model.likelihood(subseq)
            viterbi_score1 = alt_result.loglikelihood
            score = viterbi_score1 - viterbi_score0
            window = Interval(subseq.start, subseq.start + len(subseq))
            search_results.append(
                score, window, alt_result.path, viterbi_score1, viterbi_score0
            )

        return search_results


def create_profile(
    hmm: HMMERModel,
    base_abc: BaseAlphabet,
    window_length: int = 0,
    epsilon: float = 0.1,
) -> ProteinProfile:

    amino_abc = hmm.alphabet

    lprobs = lprob_normalize(hmm.insert_lprobs(0))
    null_aminot = AminoTable.create(amino_abc, lprobs)
    factory = ProteinStateFactory(CodonTable(base_abc, amino_abc), epsilon)

    nodes: List[ProteinNode] = []
    for m in range(1, hmm.model_length + 1):
        lprobs = lprob_normalize(hmm.match_lprobs(m))
        M = factory.create(f"M{m}".encode(), AminoTable.create(amino_abc, lprobs))

        lprobs = lprob_normalize(hmm.insert_lprobs(m))
        I = factory.create(f"I{m}".encode(), AminoTable.create(amino_abc, lprobs))

        D = MuteState.create(f"D{m}".encode(), base_abc)

        nodes.append(ProteinNode(M, I, D))

    trans: List[Transitions] = []
    for t in hmm.transitions:
        t.normalize()
        trans.append(t)

    profid = ProfileID(hmm.model_id.name, hmm.model_id.acc)
    prof = ProteinProfile.create(
        profid, factory, null_aminot, nodes, trans, EntryDistr.UNIFORM
    )
    prof.window_length = window_length
    return prof


def create_profile2(
    hmm: HMMERModel,
    base_abc: BaseAlphabet,
    window_length: int = 0,
    epsilon: float = 0.1,
) -> ProteinProfile:

    amino_abc = hmm.alphabet

    null_lprobs = hmm.null_lprobs
    null_log_odds = [0.0] * len(null_lprobs)

    null_aminot = AminoTable.create(amino_abc, null_lprobs)
    factory = ProteinStateFactory(CodonTable(base_abc, amino_abc), epsilon)

    nodes: List[ProteinNode] = []
    for m in range(1, hmm.model_length + 1):
        lodds = [v0 - v1 for v0, v1 in zip(hmm.match_lprobs(m), null_lprobs)]
        M = factory.create(f"M{m}".encode(), AminoTable.create(amino_abc, lodds))

        I = factory.create(
            f"I{m}".encode(), AminoTable.create(amino_abc, null_log_odds)
        )

        D = MuteState.create(f"D{m}".encode(), base_abc)

        nodes.append(ProteinNode(M, I, D))

    trans = hmm.transitions

    profid = ProfileID(hmm.model_id.name, hmm.model_id.acc)
    entry_distr = EntryDistr.OCCUPANCY
    prof = ProteinProfile.create(
        profid, factory, null_aminot, nodes, trans, entry_distr
    )
    prof.window_length = window_length
    return prof


class ProteinStateFactory:
    def __init__(
        self, gcode: CodonTable, epsilon: float,
    ):
        self._gcode = gcode
        self._epsilon = epsilon

    def create(self, name: bytes, aminot: AminoTable) -> FrameState:
        codonp = _create_codon_prob(aminot, self._gcode)
        baset = _create_base_table(codonp)
        codont = nmm.CodonTable.create(codonp)
        return FrameState.create(name, baset, codont, self._epsilon)

    @property
    def genetic_code(self) -> CodonTable:
        return self._gcode

    @property
    def epsilon(self) -> float:
        return self._epsilon


def _create_base_table(codonp: CodonProb):
    base_abc = codonp.alphabet
    base_lprob = {base: lprob_zero() for base in base_abc.symbols}
    norm = log(3)
    for codon in codon_iter(base_abc):
        lprob = codonp.get_lprob(codon)
        triplet = codon.symbols

        base_lprob[triplet[0]] = lprob_add(base_lprob[triplet[0]], lprob - norm)
        base_lprob[triplet[1]] = lprob_add(base_lprob[triplet[1]], lprob - norm)
        base_lprob[triplet[2]] = lprob_add(base_lprob[triplet[2]], lprob - norm)

    return BaseTable.create(base_abc, [base_lprob[base] for base in base_abc.symbols])


def _create_codon_prob(aminot: AminoTable, gencode: CodonTable) -> CodonProb:
    codonp = CodonProb.create(gencode.base_alphabet)

    codon_lprobs = []
    lprob_norm = lprob_zero()
    for i in range(len(aminot.alphabet.symbols)):
        aa = aminot.alphabet.symbols[i : i + 1]
        lprob = aminot.lprob(aa)

        codons = gencode.codons(aa)
        if len(codons) == 0:
            continue

        norm = log(len(codons))
        for codon in codons:
            codon_lprobs.append((codon, lprob - norm))
            lprob_norm = lprob_add(lprob_norm, codon_lprobs[-1][1])

    for codon, lprob in codon_lprobs:
        codonp.set_lprob(codon, lprob - lprob_norm)

    return codonp



import os
import stat
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import Mapping

__all__ = [
    "file_hash",
    "make_sure_dir_exist",
    "brotli_decompress",
    "tmp_cwd",
    "diff",
    "fetch_file",
    "make_executable",
    "assert_file_hash",
]


def file_hash(filepath: Path) -> str:
    import hashlib

    BLOCK_SIZE = 65536

    sha256 = hashlib.sha256()
    with open(filepath, "rb") as f:
        fb = f.read(BLOCK_SIZE)
        while len(fb) > 0:
            sha256.update(fb)
            fb = f.read(BLOCK_SIZE)

    return sha256.hexdigest()


def make_sure_dir_exist(dirpath: Path):
    dirpath.mkdir(parents=True, exist_ok=True)


def brotli_decompress(filepath: Path):
    """
    Decompress a brotli file.
    """
    from brotli import decompress

    if filepath.suffix != ".br":
        raise ValueError("File suffix must be `.br`.")

    output_filepath = filepath.parents[0] / filepath.stem
    if output_filepath.exists():
        raise RuntimeError(f"`{output_filepath}` already exists.")

    with open(filepath, "rb") as input:
        with open(output_filepath, "wb") as output:
            output.write(decompress(input.read()))

    return output_filepath


@contextmanager
def tmp_cwd():
    """
    Create and enter a temporary directory.
    The previous working directory is saved and switched back when
    leaving the context. The temporary directory is also recursively
    removed at the context ending.
    """
    oldpwd = os.getcwd()
    with tempfile.TemporaryDirectory() as tmpdir:

        os.chdir(tmpdir)
        try:
            yield
        finally:
            os.chdir(oldpwd)


def diff(filepath_a, filepath_b):
    """
    Difference between two files.
    """
    from difflib import ndiff

    with open(filepath_a, "r") as file_a:
        a = file_a.readlines()

    with open(filepath_b, "r") as file_b:
        b = file_b.readlines()

    return "".join([line.expandtabs(1) for line in ndiff(a, b)])


def make_executable(filepath):
    st = os.stat(filepath)
    os.chmod(filepath, st.st_mode | stat.S_IEXEC)


def fetch_file(
    filename: str, cache_subdir: str, url_base: str, filemap: Mapping[str, str]
) -> Path:
    import requests

    from .environment import ISEQ_CACHE_HOME

    filepath = ISEQ_CACHE_HOME / cache_subdir / filename

    if filename + ".br" in filemap:
        zipped = fetch_file(filename + ".br", cache_subdir, url_base, filemap)
        cleanup_invalid_filepath(filepath, filemap[filepath.name])
        if not filepath.exists():
            brotli_decompress(zipped)
    else:
        if filename not in filemap:
            raise ValueError(f"Unknown filename {filename}.")

        cleanup_invalid_filepath(filepath, filemap[filepath.name])

    if not filepath.exists():
        r = requests.get(f"{url_base}/{filename}")
        r.raise_for_status()
        with open(filepath, "wb") as f:
            f.write(r.content)

        assert_file_hash(filepath, filemap[filename])

    return filepath


def assert_file_hash(filepath: Path, filehash: str):
    if file_hash(filepath) != filehash:
        msg = (
            f"Hash mismatch:\n"
            f"  ACTUAL : {file_hash(filepath)}\n"
            f"  DESIRED: {filehash}"
        )
        raise RuntimeError(msg)


def cleanup_invalid_filepath(filepath: Path, filehash: str):
    if filepath.exists() and file_hash(filepath) != filehash:
        filepath.unlink()



import pytest

from iseq.gencode import GeneticCode


def test_genetic_code():
    with pytest.raises(ValueError):
        GeneticCode("Nao sei")

    with pytest.raises(ValueError):
        GeneticCode("Nao sei", "Sei la")

    gcode = GeneticCode("Invertebrate Mitochondrial")
    assert gcode.name == "Invertebrate Mitochondrial"
    assert gcode.alt_name == "SGC4"
    assert gcode.id == 5

    gcode = GeneticCode(alt_name="SGC4")
    assert gcode.name == "Invertebrate Mitochondrial"
    assert gcode.alt_name == "SGC4"
    assert gcode.id == 5

    gcode = GeneticCode(id=5)
    assert gcode.name == "Invertebrate Mitochondrial"
    assert gcode.alt_name == "SGC4"
    assert gcode.id == 5

    with pytest.raises(ValueError):
        GeneticCode(alt_name="")

    gcode = GeneticCode(name="Cephalodiscidae Mitochondrial")
    assert gcode.name == "Cephalodiscidae Mitochondrial"
    assert gcode.alt_name == ""
    assert gcode.id == 33



from math import log
from typing import List, Tuple

from imm import MuteState, NormalState, Sequence, SequenceABC, lprob_zero
from nmm import BaseAlphabet, Codon, CodonState

from .codon_table import CodonTable
from .fragment import Fragment
from .typing import AminoFragment, AminoPath, AminoStep

__all__ = ["CodonFragment"]


class CodonFragment(Fragment[BaseAlphabet, CodonState]):
    def decode(self, genetic_code: CodonTable) -> AminoFragment:
        aminos: List[bytes] = []
        steps: List[AminoStep] = []

        amino_abc = genetic_code.amino_alphabet

        start = 0
        for step in self.path:
            if isinstance(step.state, MuteState):

                mstate = MuteState.create(step.state.name, amino_abc)
                steps.append(AminoStep.create(mstate, 0))

            elif isinstance(step.state, CodonState):

                subseq = self.sequence[start : start + step.seq_len]
                name = step.state.name

                amino, new_step = create_step(genetic_code, subseq, name)
                aminos.append(amino)
                steps.append(new_step)

            else:
                raise RuntimeError("Wrong state type.")

            start += step.seq_len

        sequence = Sequence[BaseAlphabet].create(b"".join(aminos), amino_abc)
        return AminoFragment(sequence, AminoPath.create(steps), self.homologous)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}:{str(self)}>"


def create_step(
    gcode: CodonTable, seq: SequenceABC[BaseAlphabet], state_name: bytes
) -> Tuple[bytes, AminoStep]:
    base_abc = seq.alphabet
    amino = gcode.amino_acid(Codon.create(bytes(seq), base_abc))

    amino_abc = gcode.amino_alphabet
    lprobs = [lprob_zero()] * amino_abc.length
    lprobs[amino_abc.symbol_idx(amino)] = log(1.0)
    state = NormalState.create(state_name, amino_abc, lprobs)

    return amino, AminoStep.create(state, 1)



import click

from iseq.gff import read as read_gff


@click.command()
@click.argument("gff_file", type=click.File("r"))
@click.option(
    "--output",
    type=click.Path(exists=False, dir_okay=False, writable=True, resolve_path=True),
    help="Save figure to OUTPUT.",
    default=None,
)
@click.option(
    "--quiet/--no-quiet", "-q/-nq", help="Disable verbosity.", default=False,
)
def e_value_dist(gff_file, output, quiet: bool):
    """
    Plot E-values distribution.
    """
    import seaborn as sns
    from matplotlib import pyplot as plt
    from numpy import finfo, log10

    gff = read_gff(gff_file, verbose=not quiet)
    df = gff.to_dataframe()

    if "att_E-value" not in df.columns:
        click.echo("No E-value found.")
        return

    df = df[~df["att_E-value"].isna()]

    if len(df) == 0:
        click.echo("Only NaN E-values.")
        return

    min_possible = 2 ** finfo(float).minexp
    df.loc[df["att_E-value"] == 0.0, "att_E-value"] = min_possible
    df["-log10(E-value)"] = -log10(df["att_E-value"])

    sns.set(color_codes=True)

    ax = df["-log10(E-value)"].plot(kind="hist", logx=False, logy=True)
    ax.set_title("Distribution of E-values")
    ax.set_xlabel("E-value (-log10(x))")

    if not quiet:
        click.echo("Plotting... ", nl=False)
    if output is None:
        plt.show()
    else:
        plt.savefig(output)

    if not quiet:
        click.echo("done.", nl=True)
        if output is not None:
            click.echo(f"Saved to {output}.")
